@@ -97,10 +97,10 @@ partial_call(partialobject *pto, PyObject *args, PyObject *kw)
 	assert (PyTuple_Check(pto->args));
 	assert (pto->kw == Py_None  ||  PyDict_Check(pto->kw));
 
-	if (PyTuple_GET_SIZE(pto->args) == 0) {
+	if (0 == PyTuple_GET_SIZE(pto->args)) {
 		argappl = args;
 		Py_INCREF(args);
-	} else if (PyTuple_GET_SIZE(args) == 0) {
+	} else if (0 == PyTuple_GET_SIZE(args)) {
 		argappl = pto->args;
 		Py_INCREF(pto->args);
 	} else {
@@ -119,7 +119,7 @@ partial_call(partialobject *pto, PyObject *args, PyObject *kw)
 			return NULL;
 		}
 		if (kw != NULL) {
-			if (PyDict_Merge(kwappl, kw, 1) != 0) {
+			if (PyDict_Merge(kwappl, kw, 1)) {
 				Py_DECREF(argappl);
 				Py_DECREF(kwappl);
 				return NULL;
@@ -554,6 +554,8 @@ typedef struct {
 	PyObject_HEAD
 	PyObject *inner_func;
 	PyObject *outer_func;
+    PyObject *weakreflist; /* List of weak references */
+    int unpack;
 } composeobject;
 
 static PyTypeObject compose_type;
@@ -561,33 +563,34 @@ static PyTypeObject compose_type;
 static PyObject *
 compose_new(PyTypeObject *type, PyObject *args, PyObject *kw)
 {
-	PyObject *inner_func, *outer_func;
+	PyObject *inner_func, *outer_func, *unpack = Py_False;
 	composeobject *compo;
+    
+    static char *kwlist[] = {"outer", "inner", "unpack", NULL};
+    
+    if(!PyArg_ParseTupleAndKeywords(args, kw, "OO|O", kwlist,
+                &outer_func, &inner_func, &unpack))
+        return NULL;
 
-	if (PyTuple_GET_SIZE(args) != 2) {
-		PyErr_SetString(PyExc_TypeError,
-				"type 'compose' takes exactly 2 arguments");
-		return NULL;
-	}
-	
-	outer_func = PyTuple_GET_ITEM(args, 0);
 	if (!PyCallable_Check(outer_func)) {
 		PyErr_SetString(PyExc_TypeError,
 				"all arguments to compose must be callable");
 		return NULL;
 	}
 	
-	inner_func = PyTuple_GET_ITEM(args, 1);
 	if (!PyCallable_Check(inner_func)) {
 		PyErr_SetString(PyExc_TypeError,
 				"all arguments to compose must be callable");
 		return NULL;
 	}
-
+    
 	/* create flipobject structure */
 	compo = (composeobject *)type->tp_alloc(type, 0);
 	if (compo == NULL)
 		return NULL;
+        
+    compo->unpack = PyObject_IsTrue(unpack) ? 1 : 0;
+    Py_DECREF(unpack);
 
 	compo->inner_func = inner_func;
 	Py_INCREF(inner_func);
@@ -608,12 +611,19 @@ compose_call(composeobject *compo, PyObject *args, PyObject *kw)
 		return NULL;
 	}
 	
-	inner_tuple = PyTuple_New(1);
-	/* steals the reference to inner_ret_val */
-	PyTuple_SET_ITEM(inner_tuple, 0, inner_ret_val);
+    if (compo->unpack) {
+        ret_val = PyObject_CallObject(compo->outer_func, inner_ret_val);
+        Py_DECREF(inner_ret_val);
+    }
+    else {
+	    inner_tuple = PyTuple_New(1);
+	    /* steals the reference to inner_ret_val */
+	    PyTuple_SET_ITEM(inner_tuple, 0, inner_ret_val);
 	
-	ret_val = PyObject_CallObject(compo->outer_func, inner_tuple);
-	Py_DECREF(inner_tuple);
+	    ret_val = PyObject_CallObject(compo->outer_func, inner_tuple);
+        Py_DECREF(inner_tuple);
+    }    
+        
 	return ret_val;
 }
 
@@ -636,6 +646,8 @@ static void
 compose_dealloc(composeobject *compo)
 {
 	PyObject_GC_UnTrack(compo);
+    if (compo->weakreflist != NULL)
+		PyObject_ClearWeakRefs((PyObject *) compo);
 	
 	Py_XDECREF(compo->inner_func);
 	Py_XDECREF(compo->outer_func);
@@ -666,12 +678,12 @@ static PyTypeObject compose_type = {
 	PyObject_GenericSetAttr,	/* tp_setattro */
 	0,				/* tp_as_buffer */
 	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC |
-		Py_TPFLAGS_BASETYPE,	/* tp_flags */
+		Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_WEAKREFS,	/* tp_flags */
 	compose_doc,			/* tp_doc */
 	(traverseproc)compose_traverse,	/* tp_traverse */
 	0,				/* tp_clear */
 	0,				/* tp_richcompare */
-	0,				/* tp_weaklistoffset */
+	offsetof(composeobject, weakreflist),		/* tp_weaklistoffset */
 	0,				/* tp_iter */
 	0,				/* tp_iternext */
 	0,				/* tp_methods */
@@ -693,6 +705,7 @@ static PyTypeObject compose_type = {
 typedef struct {
 	PyObject_HEAD
 	PyObject *func;
+    PyObject *weakreflist; /* List of weak references */
 } flipobject;
 
 static PyTypeObject flip_type;
@@ -736,25 +749,12 @@ flip_call(flipobject *flipo, PyObject *args, PyObject *kw)
 	assert (PyCallable_Check(flipo->func));
 
 	args_size = PyTuple_GET_SIZE(args);
-	if (args_size < 2) {
-		PyErr_SetString(PyExc_TypeError,
-				"flip object takes at least 2 non-keyword parameters");
-		return NULL;
-	}
-	
 	flipped_args = PyTuple_New(args_size);
 	
-	item = PyTuple_GetItem(args, 1);
-	Py_INCREF(item);
-	PyTuple_SET_ITEM(flipped_args, 0, item);
-	item = PyTuple_GetItem(args, 0);
-	Py_INCREF(item);
-	PyTuple_SET_ITEM(flipped_args, 1, item);
-	
-	for(i = 2; i < args_size; i++) {
+	for(i = 0; i < args_size; i++) {
 		item = PyTuple_GetItem(args, i);
 		Py_INCREF(item);
-		PyTuple_SET_ITEM(flipped_args, i, item);
+		PyTuple_SET_ITEM(flipped_args, args_size - i - 1, item);
 	}
 
 	return_val = PyObject_Call(flipo->func, flipped_args, kw);
@@ -780,6 +780,9 @@ static void
 flip_dealloc(flipobject *flipo)
 {
 	PyObject_GC_UnTrack(flipo);
+    if (flipo->weakreflist != NULL)
+		PyObject_ClearWeakRefs((PyObject *) flipo);
+    
 	Py_XDECREF(flipo->func);
 	flipo->ob_type->tp_free(flipo);
 }
@@ -807,12 +810,12 @@ static PyTypeObject flip_type = {
 	PyObject_GenericSetAttr,	/* tp_setattro */
 	0,				/* tp_as_buffer */
 	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC |
-		Py_TPFLAGS_BASETYPE,	/* tp_flags */
+		Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_WEAKREFS,	/* tp_flags */
 	flip_doc,			/* tp_doc */
 	(traverseproc)flip_traverse,	/* tp_traverse */
 	0,				/* tp_clear */
 	0,				/* tp_richcompare */
-	0,	/* tp_weaklistoffset */
+	offsetof(flipobject, weakreflist),	/* tp_weaklistoffset */
 	0,				/* tp_iter */
 	0,				/* tp_iternext */
 	0,				/* tp_methods */
